// wordwrap.c
// Wordwrapper (preformatter) for JGRPro Data files (.dat)

#include <stdio.h>
#include <iostream.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <fstream.h>

#define LERRCHCK // activates error checking of line sizes

#define CHARSPERLINE 34
#define MAXCHARSPERLINE 1000
#define MAXFILENAMELEN 200
#define INLINEMAXLEN 50000
#define INWORDMAXLEN 200

void output(char * string, fstream * fileout);
int isdigits(char * string);
int printusage(void);

int main(int argc, char ** argv) {
int i, j, linecount, cpl, wordlen, largewordcount;
bool parseline;
fstream fi, fo;
char ifilename[MAXFILENAMELEN], ofilename[MAXFILENAMELEN], line[INLINEMAXLEN], word[INWORDMAXLEN], outline[MAXCHARSPERLINE];
#ifdef LERRCHCK
char linecheck[INLINEMAXLEN+1];
#endif

cpl = CHARSPERLINE;
ifilename[0] = ofilename[0] = line[0] = word[0] = outline[0] = wordlen = 0;

for(i = 1; i < argc; i++) {
 if(argv[i][0] == '-') {
  if(isdigits(argv[i]+1)) {
   cpl = atoi(argv[i]+1);
  } else {
   printusage();
   return(1);
  }
 } else
 if(ifilename[0] == 0) {
  strcpy(ifilename, argv[i]);
 } else {
  strcpy(ofilename, argv[i]);
 }

}

if(ifilename[0] == 0) {
 printusage();
} else {
 fi.open(ifilename, ios::in);
 if(!fi.is_open()) {
  cout << "Couldn\'t open file for reading. (" << ifilename << ")\n";
  return(1);
 }
 
 if(ofilename[0] != 0) {
  fo.open(ofilename, ios::out);
  if(!fo.is_open()) {
   cout << "Couldn\'t open file for writing. (" << ofilename << ")\n";
   return(1);
  }
 }
 linecount = 0;
 while(!fi.eof()) {

#ifdef LERRCHCK
  for(i = 0; i < INLINEMAXLEN+1; i++) // getline() don't do nulls, so zero everytime..
   linecheck[i] = 0;

  fi.getline(linecheck, INLINEMAXLEN+1);
  if(strlen(linecheck) > INLINEMAXLEN) {
   fprintf(stderr, "Error at line %d! ------\n", linecount);
   fprintf(stderr, "A line was larger than the limit of %d bytes.\n", INLINEMAXLEN);
   fprintf(stderr, ": Extra crlf's may appear in output file.\n");
   fi.unget();
  }
  
  strncpy(line, linecheck, INLINEMAXLEN);
#endif
#ifndef LERRCHCK
  for(i = 0; i < INLINEMAXLEN; i++) // getline() don't do nulls, so zero everytime..
   line[i] = 0;

  fi.getline(line, INLINEMAXLEN);
#endif

  linecount++;

  parseline = 1;
  if(line[0] == '[' || line[0] == '(') { // if line begins with [ or (, line is a menu title, so don't wrap it
   // but warn the user if it exceeds cpl
   if(strlen(line) > cpl) {
    fprintf(stderr, "Warning *near* line %d! ------\n", linecount);
    fprintf(stderr, "A menu title is longer than %d chars.\n", cpl);
   }
   output(line, &fo);
  } else {
   i = 0;
   for(i = 0; parseline; i++) {
    if(line[i] == ' ') {
        // the point of word wrapping is letting some whitespaces function as crlf
     if(strlen(outline) + strlen(word) > cpl) {
      if(strlen(word) > cpl) { // splits a word with a hyphen if it is larger than cpl
        largewordcount = cpl - strlen(outline) -1; // -1 for the hyphen position
        strncat(outline, word, largewordcount);
        outline[cpl-1] = '-';
        output(outline, &fo);
        outline[0] = 0;
        strcpy(word, word+largewordcount);
        wordlen -= largewordcount;
        i--;       // make sure the whitespace gets processed again
        continue;  // and reloop
      } else {
       output(outline, &fo);
       outline[0] = 0;
      }
     }
     word[wordlen] = line[i];
     if(wordlen < INWORDMAXLEN)
      wordlen++;
     else {
      fprintf(stderr, "Error *near* line %d! ------\n", linecount);
      fprintf(stderr, "A word was larger than the limit of %d bytes.\n", INWORDMAXLEN);
      fprintf(stderr, ": The word was not completely parsed.\n");
     }
     word[wordlen] = 0;
     strcat(outline, word);
     word[0] = wordlen = 0;
    } else if(line[i] == 0) { // equal to crlf when using getline()
     parseline = 0;
     if(strlen(outline) + strlen(word) > cpl) {
      if(strlen(word) > cpl) { // splits a word with a hyphen if it is larger than cpl
      // (and is the last word of the line) almost a copy-paste from above
        largewordcount = cpl - strlen(outline) -1; // -1 for the hyphen position
        strncat(outline, word, largewordcount);
        outline[cpl-1] = '-';
        output(outline, &fo);
        outline[0] = 0;
        strcpy(word, word+largewordcount);
        wordlen -= largewordcount;
        i--;       // make sure the crlf gets processed again
        parseline = 1;
        continue;  // and reloop
      } else {
       output(outline, &fo);
       outline[0] = 0;
       output(word, &fo);
       word[0] = wordlen = 0;
      }
     } else {
      if(i == 0) { // put chars if an empty line is found, because JGRPro ignore empty lines
       outline[0] = '.';
       for(i = 1; i < cpl-1 && i < MAXCHARSPERLINE-2; i++) { // commented out: for filling lines
        outline[i] = ' ';
       }
        outline[i] = '.';
        outline[++i] = 0;
      }
      strcat(outline, word);
      word[0] = wordlen = 0;
      output(outline, &fo);
      outline[0] = 0;
     }
    } else {
     word[wordlen] = line[i];
     if(wordlen < INWORDMAXLEN)  // the easy way of avoiding buffer overflows
      wordlen++;
     else {
      fprintf(stderr, "Error *near* line %d! ------\n", linecount);
      fprintf(stderr, "A word was larger than the limit of %d bytes.\n", INWORDMAXLEN);
      fprintf(stderr, ": The word was not completely parsed.\n");
     }
     word[wordlen] = 0;
    }
   }
  }
 }
 fi.close();
 if(fo.is_open()) {
  fo.close();
  cout << "Processing complete.\n";
 }
}

return(0);
}

void output(char * string, fstream * fileout) {

 if(fileout->is_open()) {
  fileout->write(string, strlen(string));
  fileout->write("\n", strlen("\n"));
 } else {
  cout << string << "\n";
 }
}

int isdigits(char * string) {
int i;

for(i = 0; string[i] != 0; i++) {
 if(!isdigit((int)string[i])) {
  return(0);
 }
}
return(1);
}

int printusage(void) {
 cout << "Usage:\n";
 cout << "wordwrap [-<chars per line>] <filein> [<fileout>]\n";
 cout << "  -<chars per line>: characters per line (default " << CHARSPERLINE << ").\n";
 cout << "  <filein>: name of the file for input.\n";
 cout << "  <fileout>: name of the file for output.\n";
 cout << "\n";
 cout << "Notes: \n";
 cout << "- Lines starting with '(' or '[' will not be wrapped for\n";
 cout << "  ease of use with JGRPro data files.\n";
 cout << "- For most aesthetical beauty, you should remove all unintended\n";
 cout << "  CRLFs from the input file before processing.\n";
 cout << "- Output is stdio, unless specified with <fileout>. This\n";
 cout << "  output might be altered when parsed by cmd.exe so\n";
 cout << "  use the <fileout> parameter whenever possible.\n";
return(0);
}

